@@ -246,6 +246,12 @@ vbe_dir_getfd(VRT_CTX, struct worker *wrk, VCL_BACKEND dir, struct backend *bp,
 	}
 	if (cw->cw_state != CW_BE_BUSY)
 		bp->n_conn++;
+
+	if (!VTAILQ_EMPTY(&bp->cw_head) && !BE_BUSY(bp)) {
+		/* Signal the new head of the waiting queue */
+		vbe_connwait_signal_locked(bp);
+	}
+
 	Lck_Unlock(bp->director->mtx);
 
 	if (cw->cw_state == CW_BE_BUSY) {
@@ -505,7 +511,7 @@ vbe_dir_http1pipe(VRT_CTX, VCL_BACKEND d)
 	v1a.req = ctx->req->acct.req_hdrbytes;
 	ctx->req->acct.req_hdrbytes = 0;
 
-	ctx->req->res_mode = RES_PIPE;
+	ctx->req->res_pipe = 1;
 
 	retval = SC_TX_ERROR;
 	pfd = vbe_dir_getfd(ctx, ctx->req->wrk, d, bp, 0);
@@ -621,12 +627,17 @@ static void v_matchproto_(vdi_list_f)
 vbe_list(VRT_CTX, const struct director *d, struct vsb *vsb, int pflag,
     int jflag)
 {
+	char buf[VTCP_ADDRBUFSIZE];
 	struct backend *bp;
+	struct vrt_endpoint *vep;
 
 	(void)ctx;
 
 	CHECK_OBJ_NOTNULL(d, DIRECTOR_MAGIC);
 	CAST_OBJ_NOTNULL(bp, d->priv, BACKEND_MAGIC);
+	CHECK_OBJ_NOTNULL(bp->endpoint, VRT_ENDPOINT_MAGIC);
+
+	vep = bp->endpoint;
 
 	if (bp->probe != NULL)
 		VBP_Status(vsb, bp, pflag, jflag);
@@ -638,6 +649,17 @@ vbe_list(VRT_CTX, const struct director *d, struct vsb *vsb, int pflag,
 		return;
 	else
 		VSB_cat(vsb, "0/0\thealthy");
+
+	if (jflag && pflag) {
+		if (vep->ipv4 != NULL) {
+			VTCP_name(vep->ipv4, buf, sizeof buf, NULL, 0);
+			VSB_printf(vsb, "\"ipv4\": \"%s\",\n", buf);
+		}
+		if (vep->ipv6 != NULL) {
+			VTCP_name(vep->ipv6, buf, sizeof buf, NULL, 0);
+			VSB_printf(vsb, "\"ipv6\": \"%s\",\n", buf);
+		}
+	}
 }
 
 /*--------------------------------------------------------------------
@@ -877,6 +899,7 @@ VRT_new_backend_clustered(VRT_CTX, struct vsmw_cluster *vc,
 		m = vbe_methods;
 	} else {
 		be->sick = 0;
+		be->vsc->happy = UINT64_MAX;
 		m = vbe_methods_noprobe;
 	}
 
