@@ -32,6 +32,7 @@
 #include "config.h"
 
 #include "cache_varnishd.h"
+#include "acceptor/cache_acceptor.h"
 
 #include <signal.h>
 #include <stdio.h>
@@ -129,17 +130,39 @@ THR_GetWorker(void)
 
 static pthread_key_t name_key;
 
+#if defined(HAVE_PTHREAD_SETNAME_NP) && !defined(__APPLE__) &&	\
+    !defined(__NetBSD__)
+static void
+thr_setname_generic(const char *name)
+{
+	char buf[16];
+
+	/* The Linux kernel enforces a strict limitation of 15 bytes name,
+	 * truncate the name if we would overflow it.
+	 */
+	if (strlen(name) > 15) {
+		bprintf(buf, "%.14s~", name);
+		name = buf;
+	}
+
+	//lint --e{438} Last value assigned not used
+	PTOK(pthread_setname_np(pthread_self(), name));
+}
+#endif
+
 void
 THR_SetName(const char *name)
 {
 
 	PTOK(pthread_setspecific(name_key, name));
-#if defined(__APPLE__)
+#if defined(HAVE_PTHREAD_SETNAME_NP)
+#  if defined(__APPLE__)
 	(void)pthread_setname_np(name);
-#elif defined(__NetBSD__)
+#  elif defined(__NetBSD__)
 	(void)pthread_setname_np(pthread_self(), "%s", (char *)(uintptr_t)name);
-#else
-	(void)pthread_setname_np(pthread_self(), name);
+#  else
+	thr_setname_generic(name);
+#  endif
 #endif
 }
 
@@ -366,6 +389,20 @@ cli_quit(int sig)
  * Run the child process
  */
 
+static void v_matchproto_(cli_func_t)
+ccf_child_start(struct cli *cli, const char * const *av, void *priv)
+{
+	(void)av;
+	(void)priv;
+
+	VCA_Start(cli);
+}
+
+static struct cli_proto child_cmds[] = {
+	{ CLICMD_SERVER_START,		"", ccf_child_start },
+	{ NULL }
+};
+
 void
 child_main(int sigmagic, size_t altstksz)
 {
@@ -443,8 +480,9 @@ child_main(int sigmagic, size_t altstksz)
 
 
 	CLI_AddFuncs(debug_cmds);
+	CLI_AddFuncs(child_cmds);
 
-#if WITH_PERSISTENT_STORAGE
+#ifdef WITH_PERSISTENT_STORAGE
 	/* Wait for persistent storage to load if asked to */
 	if (FEATURE(FEATURE_WAIT_SILO))
 		SMP_Ready();
