@@ -66,7 +66,8 @@
   REQ_STEP(deliver,             DELIVER,	static) \
   REQ_STEP(vclfail,             VCLFAIL,	static) \
   REQ_STEP(synth,               SYNTH,		static) \
-  REQ_STEP(transmit,            TRANSMIT,	static)
+  REQ_STEP(transmit,            TRANSMIT,	static) \
+  REQ_STEP(finish,              FINISH,		static)
 
 #define REQ_STEP(l, U, priv) \
     static req_state_f cnt_##l; \
@@ -235,7 +236,7 @@ cnt_deliver(struct worker *wrk, struct req *req)
 	req->t_resp = W_TIM_real(wrk);
 	VCL_deliver_method(req->vcl, wrk, req, NULL, NULL);
 
-	assert(req->restarts <= cache_param->max_restarts);
+	assert(req->restarts <= req->max_restarts);
 
 	if (wrk->vpi->handling != VCL_RET_DELIVER) {
 		HSH_Cancel(wrk, req->objcore, NULL);
@@ -355,8 +356,7 @@ cnt_synth(struct worker *wrk, struct req *req)
 		return (REQ_FSM_DONE);
 	}
 
-	if (wrk->vpi->handling == VCL_RET_RESTART &&
-	    req->restarts > cache_param->max_restarts)
+	if (wrk->vpi->handling == VCL_RET_RESTART && req->restarts > req->max_restarts)
 		wrk->vpi->handling = VCL_RET_DELIVER;
 
 	if (wrk->vpi->handling == VCL_RET_RESTART) {
@@ -429,7 +429,8 @@ cnt_synth(struct worker *wrk, struct req *req)
 static enum req_fsm_nxt v_matchproto_(req_state_f)
 cnt_transmit(struct worker *wrk, struct req *req)
 {
-	struct boc *boc;
+	enum req_fsm_nxt nxt = REQ_FSM_MORE;
+	enum vtr_deliver_e dnxt;
 	uint16_t status;
 	int sendbody, head;
 	intmax_t clval;
@@ -439,18 +440,20 @@ cnt_transmit(struct worker *wrk, struct req *req)
 	CHECK_OBJ_NOTNULL(req->transport, TRANSPORT_MAGIC);
 	CHECK_OBJ_NOTNULL(req->objcore, OBJCORE_MAGIC);
 	AZ(req->stale_oc);
-	AZ(req->res_mode);
+	AZ(req->res_pipe | req->res_esi);
+	AZ(req->boc);
+	req->req_step = R_STP_FINISH;
 
 	/* Grab a ref to the bo if there is one (=streaming) */
-	boc = HSH_RefBoc(req->objcore);
-	if (boc && boc->state < BOS_STREAM)
+	req->boc = HSH_RefBoc(req->objcore);
+	if (req->boc && req->boc->state < BOS_STREAM)
 		ObjWaitState(req->objcore, BOS_STREAM);
 	clval = http_GetContentLength(req->resp);
 	/* RFC 7230, 3.3.3 */
 	status = http_GetStatus(req->resp);
 	head = http_method_eq(req->http0->hd[HTTP_HDR_METHOD].b, HEAD);
 
-	if (boc != NULL || (req->objcore->flags & (OC_F_FAILED)))
+	if (req->boc != NULL || (req->objcore->flags & (OC_F_FAILED)))
 		req->resp_len = clval;
 	else
 		req->resp_len = ObjGetLen(req->wrk, req->objcore);
@@ -470,7 +473,7 @@ cnt_transmit(struct worker *wrk, struct req *req)
 		VSLb(req->vsl, SLT_Error, "Failure to push processors");
 		req->doclose = SC_OVERLOAD;
 		req->acct.resp_bodybytes +=
-			VDP_Close(req->vdc, req->objcore, boc);
+			VDP_Close(req->vdc, req->objcore, req->boc);
 	} else {
 		if (status < 200 || status == 204) {
 			// rfc7230,l,1691,1695
@@ -497,8 +500,21 @@ cnt_transmit(struct worker *wrk, struct req *req)
 		}
 		if (req->resp_len == 0)
 			sendbody = 0;
-		req->transport->deliver(req, boc, sendbody);
+		dnxt = req->transport->deliver(req, sendbody);
+		if (dnxt == VTR_D_DISEMBARK)
+			nxt = REQ_FSM_DISEMBARK;
+		else
+			assert(dnxt == VTR_D_DONE);
 	}
+	return (nxt);
+}
+
+static enum req_fsm_nxt v_matchproto_(req_state_f)
+cnt_finish(struct worker *wrk, struct req *req)
+{
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 
 	VSLb_ts_req(req, "Resp", W_TIM_real(wrk));
 
@@ -508,14 +524,17 @@ cnt_transmit(struct worker *wrk, struct req *req)
 		req->doclose = SC_TX_ERROR;
 	}
 
-	if (boc != NULL)
+	if (req->boc != NULL) {
 		HSH_DerefBoc(wrk, req->objcore);
+		req->boc = NULL;
+	}
 
 	(void)HSH_DerefObjCore(wrk, &req->objcore, HSH_RUSH_POLICY);
 	http_Teardown(req->resp);
 
 	req->vdp_filter_list = NULL;
-	req->res_mode = 0;
+	req->res_pipe = 0;
+	req->res_esi = 0;
 	return (REQ_FSM_DONE);
 }
 
@@ -851,7 +870,7 @@ cnt_restart(struct worker *wrk, struct req *req)
 	AZ(req->objcore);
 	AZ(req->stale_oc);
 
-	if (++req->restarts > cache_param->max_restarts) {
+	if (++req->restarts > req->max_restarts) {
 		VSLb(req->vsl, SLT_VCL_Error, "Too many restarts");
 		req->err_code = 503;
 		req->req_step = R_STP_SYNTH;
@@ -909,6 +928,8 @@ cnt_recv_prep(struct req *req, const char *ci)
 	req->is_hitpass = 0;
 	req->err_code = 0;
 	req->err_reason = NULL;
+
+	req->vfp_filter_list = NULL;
 }
 
 /*--------------------------------------------------------------------
@@ -941,6 +962,7 @@ cnt_recv(struct worker *wrk, struct req *req)
 	if (http_CountHdr(req->http0, H_Host) > 1) {
 		VSLb(req->vsl, SLT_BogoHeader, "Multiple Host: headers");
 		wrk->stats->client_req_400++;
+		req->doclose = SC_RX_BAD;
 		(void)req->transport->minimal_response(req, 400);
 		return (REQ_FSM_DONE);
 	}
@@ -948,6 +970,7 @@ cnt_recv(struct worker *wrk, struct req *req)
 	if (http_CountHdr(req->http0, H_Content_Length) > 1) {
 		VSLb(req->vsl, SLT_BogoHeader, "Multiple Content-Length: headers");
 		wrk->stats->client_req_400++;
+		req->doclose = SC_RX_BAD;
 		(void)req->transport->minimal_response(req, 400);
 		return (REQ_FSM_DONE);
 	}
@@ -1165,6 +1188,7 @@ CNT_Request(struct req *req)
 	 */
 	assert(
 	    req->req_step == R_STP_LOOKUP ||
+	    req->req_step == R_STP_FINISH ||
 	    req->req_step == R_STP_TRANSPORT);
 
 	AN(VXID_TAG(req->vsl->wid) & VSL_CLIENTMARKER);
