@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -33,12 +35,18 @@
 #include <stdio.h>
 
 #include "cache_varnishd.h"
+#include "cache_objhead.h"
+#include "cache_transport.h"
 #include "common/heritage.h"
 
 #include "vcl.h"
+#include "vtim.h"
+#include "vtcp.h"
 
 #include "vrt_obj.h"
 
+#define VRT_TMO(tmo) (isinf(tmo) ? VRT_DECIMAL_MAX : tmo)
+
 static char vrt_hostname[255] = "";
 
 /*--------------------------------------------------------------------
@@ -46,14 +54,14 @@ static char vrt_hostname[255] = "";
  */
 
 static void
-vrt_do_string(VRT_CTX, struct http *hp, int fld,
-    const char *err, const char *p, va_list ap)
+vrt_do_strands(VRT_CTX, struct http *hp, int fld,
+    const char *err, const char *str, VCL_STRANDS s)
 {
 	const char *b;
 
 	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
 
-	b = VRT_String(hp->ws, NULL, p, ap);
+	b = VRT_StrandsWS(hp->ws, str, s);
 	if (b == NULL) {
 		VRT_fail(ctx, "Workspace overflow (%s)", err);
 		WS_MarkOverflow(hp->ws);
@@ -68,14 +76,11 @@ vrt_do_string(VRT_CTX, struct http *hp, int fld,
 
 #define VRT_HDR_L(obj, hdr, fld)					\
 VCL_VOID								\
-VRT_l_##obj##_##hdr(VRT_CTX, const char *p, ...)			\
+VRT_l_##obj##_##hdr(VRT_CTX, const char *str, VCL_STRANDS s)		\
 {									\
-	va_list ap;							\
 									\
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
-	va_start(ap, p);						\
-	vrt_do_string(ctx, ctx->http_##obj, fld, #obj "." #hdr, p, ap);	\
-	va_end(ap);							\
+	vrt_do_strands(ctx, ctx->http_##obj, fld, #obj "." #hdr, str, s);	\
 }
 
 #define VRT_HDR_R(obj, hdr, fld)					\
@@ -108,7 +113,7 @@ VRT_l_##obj##_status(VRT_CTX, VCL_INT num)				\
 		VRT_fail(ctx, "illegal %s.status (%jd) (..0##)",	\
 		    #obj, (intmax_t)num);				\
 	else								\
-		http_SetStatus(ctx->http_##obj, (uint16_t)num);		\
+		http_SetStatus(ctx->http_##obj, (uint16_t)num, NULL);	\
 }
 
 #define VRT_STATUS_R(obj)						\
@@ -179,50 +184,94 @@ VRT_r_obj_reason(VRT_CTX)
 }
 
 /*--------------------------------------------------------------------
- * bool-fields (.do_*)
+ * beresp bool-fields
  */
 
-#define VBERESPW0(field)
-#define VBERESPW1(field)						\
+static inline int
+beresp_filter_fixed(VRT_CTX, const char *s)
+{
+	if (ctx->bo->vfp_filter_list == NULL)
+		return (0);
+	VRT_fail(ctx,
+	    "beresp.filters are already fixed, beresp.%s is undefined", s);
+	return (1);
+}
+
+#define VBERESPWF0(ctx, str) (void) 0
+#define VBERESPWF1(ctx, str) do {		\
+	if (beresp_filter_fixed((ctx), str))	\
+		return;			\
+	} while(0)
+
+#define VBERESPW0(field, str, fltchk)
+#define VBERESPW1(field, str, fltchk)					\
 void									\
 VRT_l_beresp_##field(VRT_CTX, VCL_BOOL a)				\
 {									\
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);			\
+	VBERESPWF##fltchk(ctx, str);					\
 	ctx->bo->field = a ? 1 : 0;					\
 }
 
-#define VBERESPR0(field)
-#define VBERESPR1(field)						\
+#define VBERESPRF0(ctx, str) (void) 0
+#define VBERESPRF1(ctx, str) do {		\
+	if (beresp_filter_fixed((ctx), str))	\
+		return (0);			\
+	} while(0)
+
+#define VBERESPR1(field, str, fltchk)					\
 VCL_BOOL								\
 VRT_r_beresp_##field(VRT_CTX)						\
+{									\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
+	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);			\
+	VBERESPRF##fltchk(ctx, str);					\
+	return (ctx->bo->field);					\
+}
+
+#define BERESP_FLAG(l, r, w, f, d)		\
+	VBERESPR##r(l, #l, f)			\
+	VBERESPW##w(l, #l, f)
+#include "tbl/beresp_flags.h"
+
+#undef VBERESPWF0
+#undef VBERESPWF1
+#undef VBERESPW0
+#undef VBERESPW1
+
+#undef VBERESPRF0
+#undef VBERESPRF1
+#undef VBERESPR1
+
+/*--------------------------------------------------------------------
+ * bereq bool-fields
+ */
+
+#define VBEREQR0(field, str)
+#define VBEREQR1(field, str)						\
+VCL_BOOL								\
+VRT_r_bereq_##field(VRT_CTX)						\
 {									\
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);			\
 	return (ctx->bo->field);					\
 }
 
-#define BO_FLAG(l, r, w, d) \
-	VBERESPR##r(l) \
-	VBERESPW##w(l)
-#include "tbl/bo_flags.h"
+#define BEREQ_FLAG(l, r, w, d)		\
+	VBEREQR##r(l, #l)
+#include "tbl/bereq_flags.h"
 
+#undef VBEREQR0
+#undef VBEREQR1
 /*--------------------------------------------------------------------*/
 
-VCL_BOOL
-VRT_r_bereq_is_bgfetch(VRT_CTX)
-{
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-	return (ctx->bo->is_bgfetch);
-}
-
 VCL_BOOL
 VRT_r_bereq_uncacheable(VRT_CTX)
 {
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-	return (ctx->bo->do_pass);
+	return (ctx->bo->uncacheable);
 }
 
 VCL_VOID
@@ -255,29 +304,81 @@ VRT_r_beresp_uncacheable(VRT_CTX)
 	return (ctx->bo->uncacheable);
 }
 
+VCL_VOID
+VRT_l_req_trace(VRT_CTX, VCL_BOOL a)
+{
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
+	ctx->req->trace = a;
+	VRT_trace(ctx, a);
+}
+VCL_VOID
+VRT_l_bereq_trace(VRT_CTX, VCL_BOOL a)
+{
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+	ctx->bo->trace = a;
+	VRT_trace(ctx, a);
+}
+
+/*--------------------------------------------------------------------*/
+
+VCL_BYTES
+VRT_r_beresp_transit_buffer(VRT_CTX)
+{
+    struct objcore *oc;
+
+    CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+    CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+
+    oc = ctx->bo->fetch_objcore;
+    CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+
+    return oc->boc->transit_buffer;
+}
+
+VCL_VOID
+VRT_l_beresp_transit_buffer(VRT_CTX, VCL_BYTES value)
+{
+    struct objcore *oc;
+
+    CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+    CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+
+    oc = ctx->bo->fetch_objcore;
+    CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+
+    oc->boc->transit_buffer = value;
+}
+
 /*--------------------------------------------------------------------*/
 
 VCL_STRING
 VRT_r_client_identity(VRT_CTX)
 {
+	const char *id;
+
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	if (ctx->req->client_identity != NULL)
-		return (ctx->req->client_identity);
-	return (SES_Get_String_Attr(ctx->req->sp, SA_CLIENT_IP));
+	if (ctx->req != NULL) {
+		CHECK_OBJ(ctx->req, REQ_MAGIC);
+		id = ctx->req->client_identity;
+	} else {
+		CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+		id = ctx->bo->client_identity;
+	}
+	if (id != NULL)
+		return (id);
+	return (SES_Get_String_Attr(ctx->sp, SA_CLIENT_IP));
 }
 
 VCL_VOID
-VRT_l_client_identity(VRT_CTX, const char *str, ...)
+VRT_l_client_identity(VRT_CTX, const char *str, VCL_STRANDS s)
 {
-	va_list ap;
 	const char *b;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	va_start(ap, str);
-	b = VRT_String(ctx->req->http->ws, NULL, str, ap);
-	va_end(ap);
+	b = VRT_StrandsWS(ctx->req->http->ws, str, s);
 	if (b == NULL) {
 		VSLb(ctx->vsl, SLT_LostHeader, "client.identity");
 		WS_MarkOverflow(ctx->req->http->ws);
@@ -288,28 +389,41 @@ VRT_l_client_identity(VRT_CTX, const char *str, ...)
 
 /*--------------------------------------------------------------------*/
 
-#define BEREQ_TIMEOUT(which)					\
+#define BEREQ_TIMEOUT(prefix, which)				\
 VCL_VOID							\
 VRT_l_bereq_##which(VRT_CTX, VCL_DURATION num)			\
 {								\
 								\
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);		\
-	ctx->bo->which = (num > 0.0 ? num : 0.0);		\
+	ctx->bo->which = num;					\
 }								\
 								\
 VCL_DURATION							\
 VRT_r_bereq_##which(VRT_CTX)					\
 {								\
+	vtim_dur res;						\
 								\
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);		\
-	return (ctx->bo->which);				\
+	res = BUSYOBJ_TMO(ctx->bo, prefix, which);		\
+	return (VRT_TMO(res));					\
+}								\
+								\
+VCL_VOID							\
+VRT_u_bereq_##which(VRT_CTX)					\
+{								\
+								\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
+	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);		\
+	ctx->bo->which = NAN;					\
 }
 
-BEREQ_TIMEOUT(connect_timeout)
-BEREQ_TIMEOUT(first_byte_timeout)
-BEREQ_TIMEOUT(between_bytes_timeout)
+BEREQ_TIMEOUT(, connect_timeout)
+BEREQ_TIMEOUT(, first_byte_timeout)
+BEREQ_TIMEOUT(, between_bytes_timeout)
+BEREQ_TIMEOUT(pipe_, task_deadline)
+
 
 /*--------------------------------------------------------------------*/
 
@@ -393,24 +507,18 @@ VRT_r_beresp_storage_hint(VRT_CTX)
 }
 
 VCL_VOID
-VRT_l_beresp_storage_hint(VRT_CTX, const char *str, ...)
+VRT_l_beresp_storage_hint(VRT_CTX, const char *str, VCL_STRANDS s)
 {
 	const char *p;
-	va_list ap;
-	uintptr_t sn;
 	VCL_STEVEDORE stv;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
 
-	sn = WS_Snapshot(ctx->ws);
-	va_start(ap, str);
-	p = VRT_String(ctx->ws, NULL, str, ap);
-	va_end(ap);
+	p = VRT_StrandsWS(ctx->ws, str, s);
 
 	if (p == NULL) {
 		VSLb(ctx->vsl, SLT_LostHeader, "storage_hint");
-		WS_Reset(ctx->ws, sn);
 		WS_MarkOverflow(ctx->ws);
 		return;
 	}
@@ -418,8 +526,6 @@ VRT_l_beresp_storage_hint(VRT_CTX, const char *str, ...)
 	stv = VRT_stevedore(p);
 	if (stv != NULL)
 		ctx->bo->storage = stv;
-
-	WS_Reset(ctx->ws, sn);
 }
 
 /*--------------------------------------------------------------------*/
@@ -429,11 +535,26 @@ VRT_r_obj_storage(VRT_CTX)
 {
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
+	AN(ctx->req->objcore->stobj);
+	CHECK_OBJ_NOTNULL(ctx->req->objcore->stobj->stevedore,
+	    STEVEDORE_MAGIC);
 	return (ctx->req->objcore->stobj->stevedore);
 }
 
 /*--------------------------------------------------------------------*/
 
+VCL_BOOL
+VRT_r_obj_can_esi(VRT_CTX)
+{
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
+	return (ObjHasAttr(ctx->req->wrk, ctx->req->objcore, OA_ESIDATA));
+}
+
+/*--------------------------------------------------------------------*/
+
 #define REQ_VAR_L(nm, elem, type, extra)				\
 									\
 VCL_VOID								\
@@ -455,12 +576,33 @@ VRT_r_req_##nm(VRT_CTX)							\
 	return (ctx->req->elem);					\
 }
 
-REQ_VAR_L(backend_hint, director_hint, VCL_BACKEND,)
+#define REQ_VAR_U(nm, elem, val)					\
+									\
+VCL_VOID								\
+VRT_u_req_##nm(VRT_CTX)							\
+{									\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
+	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);				\
+	ctx->req->elem = val;						\
+}
+
 REQ_VAR_R(backend_hint, director_hint, VCL_BACKEND)
+
 REQ_VAR_L(ttl, d_ttl, VCL_DURATION, if (!(arg>0.0)) arg = 0;)
 REQ_VAR_R(ttl, d_ttl, VCL_DURATION)
+REQ_VAR_U(ttl, d_ttl, -1)
 REQ_VAR_L(grace, d_grace, VCL_DURATION, if (!(arg>0.0)) arg = 0;)
 REQ_VAR_R(grace, d_grace, VCL_DURATION)
+REQ_VAR_U(grace, d_grace, -1)
+
+VCL_VOID
+VRT_l_req_backend_hint(VRT_CTX, VCL_BACKEND be)
+{
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
+	VRT_Assign_Backend(&ctx->req->director_hint, be);
+}
 
 /*--------------------------------------------------------------------*/
 
@@ -470,7 +612,7 @@ VRT_l_bereq_backend(VRT_CTX, VCL_BACKEND be)
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-	ctx->bo->director_req = be;
+	VRT_Assign_Backend(&ctx->bo->director_req, be);
 }
 
 VCL_BACKEND
@@ -498,12 +640,17 @@ VRT_u_bereq_body(VRT_CTX)
 {
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+
+	http_Unset(ctx->bo->bereq, H_Content_Length);
+
+	if (ctx->bo->bereq_body != NULL)
+		(void)HSH_DerefObjCore(ctx->bo->wrk, &ctx->bo->bereq_body, 0);
+
 	if (ctx->bo->req != NULL) {
-		CHECK_OBJ_NOTNULL(ctx->bo->req, REQ_MAGIC);
+		CHECK_OBJ(ctx->bo->req, REQ_MAGIC);
 		ctx->bo->req = NULL;
 		ObjSetState(ctx->bo->wrk,
 		    ctx->bo->fetch_objcore, BOS_REQ_DONE);
-		http_Unset(ctx->bo->bereq, H_Content_Length);
 	}
 }
 
@@ -519,8 +666,10 @@ VRT_l_req_esi(VRT_CTX, VCL_BOOL process_esi)
 	/*
 	 * Only allow you to turn of esi in the main request
 	 * else everything gets confused
+	 * NOTE: this is not true, but we do not change behavior
+	 * for vcl 4.0. For 4.1, see VRT_l_resp_do_esi()
 	 */
-	if (ctx->req->esi_level == 0)
+	if (IS_TOPREQ(ctx->req))
 		ctx->req->disable_esi = !process_esi;
 }
 
@@ -574,6 +723,18 @@ VRT_r_bereq_retries(VRT_CTX)
 	return (ctx->bo->retries);
 }
 
+/*--------------------------------------------------------------------*/
+
+VCL_STRING
+VRT_r_req_transport(VRT_CTX)
+{
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+
+	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->req->transport, TRANSPORT_MAGIC);
+	return (ctx->req->transport->name);
+}
+
 /*--------------------------------------------------------------------
  * In exp.*:
  *	t_origin is absolute
@@ -647,6 +808,25 @@ VRT_DO_EXP_R(beresp, ctx->bo->fetch_objcore, keep, 0)
 
 /*lint -restore */
 
+// XXX more assertions?
+#define VRT_DO_TIME_R(which, where, field)				\
+									\
+VCL_TIME								\
+VRT_r_##which##_time(VRT_CTX)						\
+{									\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
+	AN((ctx)->where);						\
+									\
+	return ((ctx)->where->field);					\
+}
+
+VRT_DO_TIME_R(req, req, t_req)
+VRT_DO_TIME_R(req_top, req->top->topreq, t_req)
+VRT_DO_TIME_R(resp, req, t_resp)
+VRT_DO_TIME_R(bereq, bo, t_first)
+VRT_DO_TIME_R(beresp, bo, t_resp)
+VRT_DO_TIME_R(obj, req->objcore, t_origin)
+
 /*--------------------------------------------------------------------
  */
 
@@ -667,40 +847,45 @@ VRT_DO_AGE_R(beresp, ctx->bo->fetch_objcore)
  * [[be]req|sess].xid
  */
 
-VCL_STRING
+VCL_INT
 VRT_r_req_xid(VRT_CTX)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-
-	return (WS_Printf(ctx->req->http->ws, "%u",
-	    VXID(ctx->req->vsl->wid)));
+	CHECK_OBJ_NOTNULL(ctx->req->http, HTTP_MAGIC);
+	AN(ctx->req->vsl);
+	return (VXID(ctx->req->vsl->wid));
 }
 
-VCL_STRING
+VCL_INT
 VRT_r_bereq_xid(VRT_CTX)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+	AN(ctx->bo->vsl);
 
-	return (WS_Printf(ctx->bo->bereq->ws, "%u",
-	    VXID(ctx->bo->vsl->wid)));
+	return (VXID(ctx->bo->vsl->wid));
 }
 
-VCL_STRING
+VCL_INT
 VRT_r_sess_xid(VRT_CTX)
 {
+	struct sess *sp;
+
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 
-	if (VALID_OBJ(ctx->req, REQ_MAGIC))
-		return (WS_Printf(ctx->req->http->ws, "%u",
-		    VXID(ctx->req->sp->vxid)));
+	if (ctx->req) {
+		CHECK_OBJ(ctx->req, REQ_MAGIC);
+		sp = ctx->req->sp;
+	} else {
+		CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+		sp = ctx->bo->sp;
+	}
 
-	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-	return (WS_Printf(ctx->bo->bereq->ws, "%u",
-	    VXID(ctx->bo->sp->vxid)));
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	return (VXID(sp->vxid));
 }
 
 /*--------------------------------------------------------------------
@@ -763,9 +948,10 @@ VRT_r_local_##var(VRT_CTX)					\
 	struct sess *sp;					\
 								\
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
-	if (VALID_OBJ(ctx->req, REQ_MAGIC))			\
+	if (ctx->req) {						\
+		CHECK_OBJ(ctx->req, REQ_MAGIC);			\
 		sp = ctx->req->sp;				\
-	else {							\
+	} else {						\
 		CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);	\
 		sp = ctx->bo->sp;				\
 	}							\
@@ -822,7 +1008,7 @@ VRT_r_obj_uncacheable(VRT_CTX)
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
-	return (ctx->req->objcore->flags & OC_F_PASS ? 1 : 0);
+	return (ctx->req->objcore->flags & OC_F_HFM ? 1 : 0);
 }
 
 /*--------------------------------------------------------------------*/
@@ -840,6 +1026,15 @@ VRT_r_resp_is_streaming(VRT_CTX)
 
 /*--------------------------------------------------------------------*/
 
+static inline int
+resp_filter_fixed(VRT_CTX, const char *s)
+{
+	if (ctx->req->vdp_filter_list == NULL)
+		return (0);
+	VRT_fail(ctx, "resp.filters are already fixed, %s is undefined", s);
+	return (1);
+}
+
 VCL_VOID
 VRT_l_resp_do_esi(VRT_CTX, VCL_BOOL process_esi)
 {
@@ -847,10 +1042,8 @@ VRT_l_resp_do_esi(VRT_CTX, VCL_BOOL process_esi)
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
 	assert(ctx->syntax >= 41);
-	/*
-	 * Only allow you to turn of esi in the main request
-	 * else everything gets confused
-	 */
+	if (resp_filter_fixed(ctx, "resp.do_esi"))
+		return;
 	ctx->req->disable_esi = !process_esi;
 }
 
@@ -861,6 +1054,8 @@ VRT_r_resp_do_esi(VRT_CTX)
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
 	assert(ctx->syntax >= 41);
+	if (resp_filter_fixed(ctx, "resp.do_esi"))
+		return (0);
 	return (!ctx->req->disable_esi);
 }
 
@@ -868,22 +1063,33 @@ VRT_r_resp_do_esi(VRT_CTX)
 
 #define VRT_BODY_L(which)					\
 VCL_VOID							\
-VRT_l_##which##_body(VRT_CTX, const char *str, ...)		\
+VRT_l_##which##_body(VRT_CTX, enum lbody_e type,		\
+    const char *str, VCL_BODY body)				\
 {								\
-	va_list ap;						\
-	const char *p;						\
+	int n;							\
 	struct vsb *vsb;					\
+	VCL_STRANDS s;						\
+	VCL_BLOB b;						\
 								\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
+	AN(body);						\
 	CAST_OBJ_NOTNULL(vsb, ctx->specific, VSB_MAGIC);	\
-	va_start(ap, str);					\
-	p = str;						\
-	while (p != vrt_magic_string_end) {			\
-		if (p == NULL)					\
-			p = "(null)";				\
-		VSB_cat(vsb, p);				\
-		p = va_arg(ap, const char *);			\
+	if (type == LBODY_SET_STRING || type == LBODY_SET_BLOB)	\
+		VSB_clear(vsb);					\
+	if (type == LBODY_SET_BLOB || type == LBODY_ADD_BLOB) {	\
+		AZ(str);					\
+		b = body;					\
+		VSB_bcat(vsb, b->blob, b->len);			\
+		return;						\
 	}							\
-	va_end(ap);						\
+	if (str != NULL)					\
+		VSB_cat(vsb, str);				\
+	assert(type == LBODY_SET_STRING ||			\
+	    type == LBODY_ADD_STRING);				\
+	s = body;						\
+	for (n = 0; s != NULL && n < s->n; n++)			\
+		if (s->p[n] != NULL)				\
+			VSB_cat(vsb, s->p[n]);			\
 }
 
 VRT_BODY_L(beresp)
@@ -891,12 +1097,16 @@ VRT_BODY_L(resp)
 
 /*--------------------------------------------------------------------*/
 
+			/* digest */
+#define BLOB_HASH_TYPE 0x00d16357
+
 VCL_BLOB
 VRT_r_req_hash(VRT_CTX)
 {
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	return (VRT_blob(ctx, "req.hash", ctx->req->digest, DIGEST_LEN));
+	return (VRT_blob(ctx, "req.hash", ctx->req->digest, DIGEST_LEN,
+	    BLOB_HASH_TYPE));
 }
 
 VCL_BLOB
@@ -904,7 +1114,8 @@ VRT_r_bereq_hash(VRT_CTX)
 {
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-	return (VRT_blob(ctx, "bereq.hash", ctx->bo->digest, DIGEST_LEN));
+	return (VRT_blob(ctx, "bereq.hash", ctx->bo->digest, DIGEST_LEN,
+	    BLOB_HASH_TYPE));
 }
 
 /*--------------------------------------------------------------------*/
@@ -925,32 +1136,45 @@ HTTP_VAR(beresp)
 
 /*--------------------------------------------------------------------*/
 
-VCL_STRING
-VRT_r_beresp_filters(VRT_CTX)
+static inline void
+set_idle_send_timeout(const struct sess *sp, VCL_DURATION d)
 {
-
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-	if (ctx->bo->filter_list != NULL)
-		return(ctx->bo->filter_list);
-	/* We do not set bo->filter_list yet, things might still change */
-	return (VBF_Get_Filter_List(ctx->bo));
+	struct timeval tv = VTIM_timeval_sock(d);
+	VTCP_Assert(setsockopt(sp->fd, SOL_SOCKET, SO_SNDTIMEO,
+	    &tv, sizeof tv));
 }
 
-VCL_VOID
-VRT_l_beresp_filters(VRT_CTX, const char *str, ...)
-{
-	va_list ap;
-	const char *b;
-
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-	va_start(ap, str);
-	b = VRT_String(ctx->bo->ws, NULL, str, ap);
-	va_end(ap);
-	if (b == NULL) {
-		WS_MarkOverflow(ctx->bo->ws);
-		return;
-	}
-	ctx->bo->filter_list = b;
+#define SESS_VAR_DUR(x, setter)				\
+VCL_VOID						\
+VRT_l_sess_##x(VRT_CTX, VCL_DURATION d)			\
+{							\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);		\
+	CHECK_OBJ_NOTNULL(ctx->sp, SESS_MAGIC);		\
+	d = vmax(d, 0.0);				\
+	setter;						\
+	ctx->sp->x = d;					\
+}							\
+							\
+VCL_DURATION						\
+VRT_r_sess_##x(VRT_CTX)					\
+{							\
+	vtim_dur res;					\
+							\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);		\
+	CHECK_OBJ_NOTNULL(ctx->sp, SESS_MAGIC);		\
+	res = SESS_TMO(ctx->sp, x);			\
+	return (VRT_TMO(res));				\
+}							\
+							\
+VCL_VOID						\
+VRT_u_sess_##x(VRT_CTX)					\
+{							\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);		\
+	CHECK_OBJ_NOTNULL(ctx->sp, SESS_MAGIC);		\
+	ctx->sp->x = NAN;				\
 }
+
+SESS_VAR_DUR(timeout_idle, )
+SESS_VAR_DUR(timeout_linger, )
+SESS_VAR_DUR(send_timeout, )
+SESS_VAR_DUR(idle_send_timeout, set_idle_send_timeout(ctx->sp, d))
