@@ -184,6 +184,7 @@ SML_AllocBuf(struct worker *wrk, const struct stevedore *stv, size_t size,
 	if (st == NULL)
 		return (NULL);
 	assert(st->space >= size);
+	st->flags = STORAGE_F_BUFFER;
 	st->len = size;
 	*ppriv = (uintptr_t)st;
 	return (st->ptr);
@@ -198,6 +199,7 @@ SML_FreeBuf(struct worker *wrk, const struct stevedore *stv, uintptr_t priv)
 	CHECK_OBJ_NOTNULL(stv, STEVEDORE_MAGIC);
 
 	CAST_OBJ_NOTNULL(st, (void *)priv, STORAGE_MAGIC);
+	assert(st->flags == STORAGE_F_BUFFER);
 	sml_stv_free(stv, st);
 }
 
@@ -395,15 +397,15 @@ sml_iterator(struct worker *wrk, struct objcore *oc,
 			assert(nl > 0);
 			sl += st->len;
 			st = VTAILQ_PREV(st, storagehead, list);
-			if (VTAILQ_PREV(st, storagehead, list) != NULL) {
-				if (final && checkpoint != NULL) {
-					VTAILQ_REMOVE(&obj->list,
-					    checkpoint, list);
-					sml_stv_free(stv, checkpoint);
-				}
-				checkpoint = st;
-				checkpoint_len = sl;
+			if (final && checkpoint != NULL) {
+				if (checkpoint == boc->stevedore_priv)
+					boc->stevedore_priv = trim_once;
+				else
+					VTAILQ_REMOVE(&obj->list, checkpoint, list);
+				sml_stv_free(stv, checkpoint);
 			}
+			checkpoint = st;
+			checkpoint_len = sl;
 		}
 		CHECK_OBJ_NOTNULL(obj, OBJECT_MAGIC);
 		CHECK_OBJ_NOTNULL(st, STORAGE_MAGIC);
@@ -560,7 +562,8 @@ sml_trimstore(struct worker *wrk, struct objcore *oc)
 		Lck_Lock(&oc->boc->mtx);
 		VTAILQ_REMOVE(&o->list, st, list);
 		Lck_Unlock(&oc->boc->mtx);
-		sml_stv_free(stv, st);
+		/* sml_bocdone frees this */
+		oc->boc->stevedore_priv = st;
 		return;
 	}
 
