@@ -6,4 +6,4 @@
 
 FLOPS='
 	*.c
-' ../../tools/flint_skel.sh
+' ../../tools/flint_skel.sh $*
