@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 #
 # Copyright (c) 2006 Verdens Gang AS
 # Copyright (c) 2006-2015 Varnish Software AS
@@ -6,6 +6,8 @@
 #
 # Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
 #
+# SPDX-License-Identifier: BSD-2-Clause
+#
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions
 # are met:
@@ -30,16 +32,9 @@
 # Generate various .c and .h files for the VCL compiler and the interfaces
 # for it.
 
-from __future__ import print_function
-
 #######################################################################
 # These are our tokens
 
-# We could drop all words such as "include", "if" etc, and use the
-# ID type instead, but declaring them tokens makes them reserved words
-# which hopefully makes for better error messages.
-# XXX: does it actually do that ?
-
 import copy
 import sys
 from os.path import join
@@ -78,82 +73,50 @@ tokens = {
     "CNUM":     None,
     "FNUM":     None,
     "CSTR":     None,
-    "EOI":      None,
     "CSRC":     None,
+    "CBLOB":    None,
+
+    # End of token list
+    "EOI":      None,
 }
 
 #######################################################################
 # Our methods and actions
 
-returns = (
-    ###############################################################
-    # Client side
+returns = []
 
-    ('recv',
-     "C",
-     ('fail', 'synth', 'restart', 'pass', 'pipe', 'hash', 'purge', 'vcl')
-    ),
-    ('pipe',
-     "C",
-     ('fail', 'synth', 'pipe',)
-    ),
-    ('pass',
-     "C",
-     ('fail', 'synth', 'restart', 'fetch',)
-    ),
-    ('hash',
-     "C",
-     ('fail', 'lookup',)
-    ),
-    ('purge',
-     "C",
-     ('fail', 'synth', 'restart',)
-    ),
-    ('miss',
-     "C",
-     ('fail', 'synth', 'restart', 'pass', 'fetch',)
-    ),
-    ('hit',
-     "C",
-     ('fail', 'synth', 'restart', 'pass', 'miss', 'deliver',)
-    ),
-    ('deliver',
-     "C",
-     ('fail', 'synth', 'restart', 'deliver',)
-    ),
-    ('synth',
-     "C",
-     ('fail', 'restart', 'deliver',)
-    ),
+def parse_subs(ln, c):
+    ctx = ['C', 'B', 'H']
+    rets = []
+    for i in ln[2:]:
+        if (len(i) > 4 and i.startswith('  | ``') and i.endswith('``')):
+            rets.append(i[6:-2].split('(')[0])
+    returns.append((ln[0][4:], ctx[c], rets))
 
-    ###############################################################
-    # Backend-fetch
+def parse_subs_doc(fn):
+    l = []
+    j = -1
+    start = False
+    for i in open(fn):
+        if i.startswith('##########'):
+            j += 1
+            l.append([])
+            start = True
+            continue
+        if start:
+            l[j].append(i.rstrip())
+    for i in range(0, len(l)):
+        for n in range(0, len(l[i])):
+            if len(l[i][n]) < 5 or not l[i][n].startswith("~~~~~"):
+                continue
+            m = n + 1
+            while m < len(l[i]) and (len(l[i][m]) < 5 or not l[i][m].startswith("~~~~~")):
+                m += 1
+            if m != len(l[i]):
+                m -= 2
+            parse_subs(l[i][n-1:m], i)
 
-    ('backend_fetch',
-     "B",
-     ('fail', 'fetch', 'abandon')
-    ),
-    ('backend_response',
-     "B",
-     ('fail', 'deliver', 'retry', 'abandon', 'pass')
-    ),
-    ('backend_error',
-     "B",
-     ('fail', 'deliver', 'retry', 'abandon')
-    ),
-
-    ###############################################################
-    # Housekeeping
-
-    ('init',
-     "H",
-     ('ok', 'fail')
-    ),
-    ('fini',
-     "H",
-     ('ok',)
-    ),
-)
+parse_subs_doc(join(srcroot, "doc/sphinx/reference/vcl_step.rst"))
 
 #######################################################################
 # Variables available in sessions
@@ -170,27 +133,40 @@ def varproto(s):
         fh.write(s + ";\n")
         varprotos[s] = True
 
+def var_is_wildcard(sym):
+    return sym[-2:] == '.*'
+
+def var_symbol_name(sym):
+    if var_is_wildcard(sym):
+        return sym[:-2]
+    return sym
+
 class vardef(object):
-    def __init__(self, nam, typ, rd, wr, wu, vlo, vhi):
-        self.nam = nam
+    def __init__(self, sym, typ, rd, wr, wu, al, vlo, vhi):
+        self.sym = sym
+        self.nam = var_symbol_name(sym)
         self.typ = typ
         self.rd = rd
         self.wr = wr
         self.uns = wu
+        self.al = al
         self.vlo = vlo
         self.vhi = vhi
 
-        self.emit()
+        if al is None:
+            self.emit_var()
+        else:
+            self.emit_alias()
 
-    def emit(self):
+    def emit_var(self):
         fh.write("\n")
         fo.write("\n")
         cnam = self.nam.replace(".", "_")
         ctyp = vcltypes[self.typ]
 
         # fo.write("\t{ \"%s\", %s,\n" % (nm, self.typ))
-        fo.write("\tsym = VCC_MkSym(tl, \"%s\", " % self.nam)
-        if self.typ == "HEADER":
+        fo.write("\tsym = VCC_MkSym(tl, \"%s\", SYM_MAIN," % self.nam)
+        if var_is_wildcard(self.sym):
             fo.write(" SYM_NONE, %d, %d);\n" % (self.vlo, self.vhi))
             fo.write("\tAN(sym);\n")
             fo.write("\tsym->wildcard = vcc_Var_Wildcard;\n")
@@ -198,30 +174,35 @@ class vardef(object):
             fo.write(" SYM_VAR, %d, %d);\n" % (self.vlo, self.vhi))
         fo.write("\tAN(sym);\n")
         fo.write("\tsym->type = %s;\n" % self.typ)
-        fo.write("\tsym->eval = vcc_Eval_Var;\n")
+        if self.typ == "HEADER" and not var_is_wildcard(self.sym):
+            fo.write("\tsym->eval = vcc_Eval_ProtectedHeader;\n")
+        else:
+            fo.write("\tsym->eval = vcc_Eval_Var;\n")
 
-        if self.typ == "HEADER":
+        if var_is_wildcard(self.sym):
             fo.write('\tsym->rname = "HDR_')
             fo.write(self.nam.split(".")[0].upper())
             fo.write('";\n')
-        elif self.rd:
+        elif self.rd and self.typ != "HEADER":
             fo.write('\tsym->rname = "VRT_r_%s(ctx)";\n' % cnam)
             varproto("VCL_" + self.typ + " VRT_r_%s(VRT_CTX)" % cnam)
         fo.write("\tsym->r_methods =\n")
         restrict(fo, self.rd)
         fo.write(";\n")
 
-        if self.typ == "HEADER":
+        if var_is_wildcard(self.sym):
             fo.write('\tsym->lname = "HDR_')
             fo.write(self.nam.split(".")[0].upper())
             fo.write('";\n')
         elif self.wr:
             fo.write('\tsym->lname = "VRT_l_%s(ctx, ";\n' % cnam)
             s = "void VRT_l_%s(VRT_CTX, " % cnam
-            if self.typ != "STRING" and self.typ != "BODY":
-                s += "VCL_" + self.typ + ")"
+            if self.typ == "STRING":
+                s += ctyp.c + ", VCL_STRANDS)"
+            elif self.typ == "BODY":
+                s += "enum lbody_e, const char *, " + ctyp.c + ")"
             else:
-                s += ctyp.c + ", ...)"
+                s += "VCL_" + self.typ + ")"
             varproto(s)
         fo.write("\tsym->w_methods =\n")
         restrict(fo, self.wr)
@@ -234,6 +215,11 @@ class vardef(object):
         restrict(fo, self.uns)
         fo.write(";\n")
 
+    def emit_alias(self):
+        var_aliases.append("\tsym = VCC_MkSymAlias(tl, \"%s\", \"%s\");\n" %
+                (self.nam, self.al))
+        var_aliases.append("\tAN(sym);\n")
+
 def parse_vcl(x):
     vlo, vhi = (0, 99)
     x = x.split()
@@ -251,8 +237,6 @@ def parse_var(ln):
     l1 = ln.pop(0).split("``")
     assert len(l1) in (1, 3)
     vn = l1[0].strip()
-    if vn[-1] == '*':
-        vn = vn[:-1]
     if len(l1) == 3:
         vlo, vhi = parse_vcl(l1[1])
     else:
@@ -260,6 +244,7 @@ def parse_var(ln):
     vr = []
     vw = []
     vu = []
+    va = None
     while True:
         l = ln.pop(0)
         if l == "":
@@ -277,13 +262,16 @@ def parse_var(ln):
             for i in j[2:]:
                 vw.append(i.strip(",."))
             continue
-        if j[0] == "Unsetable" and j[1] == "from:":
+        if j[0] == "Unsettable" and j[1] == "from:":
             for i in j[2:]:
                 vu.append(i.strip(",."))
             continue
+        if j[0] == "Alias" and j[1] == "of:":
+            va = j[2]
+            continue
         break
     if vn[:8] != "storage.":
-        vardef(vn, vt, vr, vw, vu, vlo, vhi)
+        vardef(vn, vt, vr, vw, vu, va, vlo, vhi)
 
 def parse_var_doc(fn):
     l = []
@@ -299,15 +287,15 @@ def parse_var_doc(fn):
         parse_var(l[n-2:m-1])
 
 stv_variables = (
-    ('free_space', 'BYTES', "0.", 'storage.<name>.free_space', """
+    ('free_space', 'BYTES', "0", 'storage.<name>.free_space', """
     Free space available in the named stevedore. Only available for
     the malloc stevedore.
     """),
-    ('used_space', 'BYTES', "0.", 'storage.<name>.used_space', """
+    ('used_space', 'BYTES', "0", 'storage.<name>.used_space', """
     Used space in the named stevedore. Only available for the malloc
     stevedore.
     """),
-    ('happy', 'BOOL', "0", 'storage.<name>.happy', """
+    ('happy', 'BOOL', "1", 'storage.<name>.happy', """
     Health status for the named stevedore. Not available in any of the
     current stevedores.
     """),
@@ -327,7 +315,6 @@ class vcltype(object):
 
 
 vcltype("STRINGS", "void", True)
-vcltype("STRING_LIST", "void*", True)
 vcltype("SUB", "void*", True)
 
 fi = open(join(srcroot, "include/vrt.h"))
@@ -431,21 +418,16 @@ def emit_vcl_tnames(fo, tokens):
 
 
 #######################################################################
-def emit_file(fo, fd, bn):
-    "Read a C-source file and spit out code that outputs it with VSB_cat()"
-    fn = join(fd, bn)
+def emit_strings(fo, name, fc):
+    "spit out code that outputs the fc iterable with VSB_cat()"
 
-    fi = open(fn)
-    fc = fi.read()
-    fi.close()
-
-    w = 66      # Width of lines, after white space prefix
+    w = 66          # Width of lines, after white space prefix
     maxlen = 10240  # Max length of string literal
 
     x = 0
     l = 0
-    fo.write("\n\t/* %s */\n\n" % fn)
-    fo.write('\tVSB_cat(sb, "/* ---===### %s ###===--- */\\n\\n");\n' % bn)
+
+    fo.write('\tVSB_cat(sb, "/* ---===### %s ###===--- */\\n\\n");\n' % name)
     for c in fc:
         if l == 0:
             fo.write("\tVSB_cat(sb, \"")
@@ -488,6 +470,18 @@ def emit_file(fo, fd, bn):
         fo.write("\t);\n")
     fo.write('\tVSB_cat(sb, "\\n");\n')
 
+
+def emit_file(fo, fd, bn):
+    "Read a C-source file and spit out code that outputs it with VSB_cat()"
+    fn = join(fd, bn)
+
+    fi = open(fn)
+    fc = fi.read()
+    fi.close()
+
+    fo.write("\n\t/* %s */\n\n" % fn)
+    emit_strings(fo, bn, fc)
+
 #######################################################################
 
 
@@ -625,6 +619,10 @@ for i in sorted(task.keys()):
     fo.write(tbl40("#define VCL_MET_TASK_%s" % i.upper(),
                    "( " + (" | \\\n\t\t\t\t\t  ").join(task[i]) + " )\n"))
 
+fo.write("\n")
+fo.write(tbl40("#define VCL_MET_TASK_ALL", "( VCL_MET_TASK_"))
+fo.write(" | \\\n\t\t\t\t\t  VCL_MET_TASK_".join(map(str.upper, task.keys())))
+fo.write(" )")
 
 fo.write("\n/* VCL Returns */\n")
 n = 1
@@ -635,6 +633,17 @@ for i in sorted(rets.keys()):
 fo.write("\n" + tbl40("#define VCL_RET_MAX", "%d\n" % n))
 
 fo.write("\n/* VCL Types */\n")
+fo.write('''
+struct vrt_type {
+\tunsigned\t\tmagic;
+#define VRT_TYPE_MAGIC\t\t0xa943bc32
+\tconst char\t\t*lname;
+\tconst char\t\t*uname;
+\tconst char\t\t*ctype;
+\tsize_t\t\t\tszof;
+};
+
+''')
 for vcltype in sorted(vcltypes.keys()):
     fo.write("extern const struct vrt_type VCL_TYPE_%s[1];\n" % vcltype)
 
@@ -644,35 +653,52 @@ fo.write("""
 typedef int vcl_event_f(VRT_CTX, enum vcl_event_e);
 typedef int vcl_init_f(VRT_CTX);
 typedef void vcl_fini_f(VRT_CTX);
-typedef void vcl_func_f(VRT_CTX);
 
 struct VCL_conf {
-    unsigned            magic;
-#define VCL_CONF_MAGIC          0x7406c509      /* from /dev/random */
+	unsigned		magic;
+#define VCL_CONF_MAGIC		0x7406c509      /* from /dev/random */
 
-    unsigned            syntax;
-    VCL_BACKEND             *default_director;
-    const struct vrt_backend_probe  *default_probe;
-    unsigned            nref;
-    const struct vrt_ref        *ref;
+	unsigned		syntax;
+	VCL_BACKEND		*default_director;
+	VCL_PROBE		default_probe;
+	unsigned		nref;
+	const struct vpi_ref	*ref;
 
-    unsigned            nsrc;
-    const char              **srcname;
-    const char              **srcbody;
+	unsigned		nsrc;
+	unsigned		nsub;
+	const char		**srcname;
+	const char		**srcbody;
 
-    unsigned            nvmod;
+	int			nvmod;
+	const struct vpi_ii	*instance_info;
 
-    vcl_event_f             *event_vcl;
+	vcl_event_f		*event_vcl;
 """)
 
 for i in returns:
-    fo.write("\tvcl_func_f\t\t\t*" + i[0] + "_func;\n")
+    fo.write("\tvcl_func_f\t\t*" + i[0] + "_func;\n")
 
 fo.write("\n};\n")
 fo.close()
 
 #######################################################################
 
+fo = open(join(buildroot, "include/tbl/vcl_context.h"), "w")
+file_header(fo)
+
+fo.write("/*lint -save -e525 -e539 */\n")
+for i in returns:
+    fo.write("\nVCL_CTX(vcl_%s,%s)" % (i[0],i[0].upper()))
+fo.write("\nVCL_CTX(backend, TASK_B)")
+fo.write("\nVCL_CTX(client, TASK_C)")
+fo.write("\nVCL_CTX(housekeeping, TASK_H)")
+fo.write("/*lint -restore */\n")
+fo.write("\n")
+fo.write("\n#undef VCL_CTX")
+fo.write("\n")
+fo.close()
+
+#######################################################################
 
 def restrict(fo, spec):
     d = dict()
@@ -733,12 +759,14 @@ vcc_Var_Init(struct vcc *tl)
     struct symbol *sym;
 """)
 
-
+var_aliases = []
 parse_var_doc(join(srcroot, "doc/sphinx/reference/vcl_var.rst"))
+for al in var_aliases:
+    fo.write(al)
 fo.write("}\n")
 
 for i in stv_variables:
-    fh.write(vcltypes[i[1]].c + " VRT_Stv_" + i[0] + "(const char *);\n")
+    fh.write("VCL_" + i[1] + " VRT_stevedore_" + i[0] + "(VCL_STEVEDORE);\n")
 
 fo.write("\n/* VCL type identifiers */\n")
 
@@ -783,18 +811,27 @@ emit_file(fo, srcroot, "include/vdef.h")
 emit_file(fo, srcroot, "include/vrt.h")
 emit_file(fo, buildroot, "include/vcl.h")
 emit_file(fo, buildroot, "include/vrt_obj.h")
+emit_file(fo, srcroot, "include/vcc_interface.h")
+emit_strings(fo, "vgc asserts (generate.py)",
+'''#define assert(e)							\\
+do {									\\
+	if (!(e)) {							\\
+		VPI_Fail(__func__, __FILE__, __LINE__, #e);		\\
+	}								\\
+} while (0)
+''')
 
 fo.write("\n}\n")
 fo.close()
 
 #######################################################################
-ft = open(join(buildroot, "include/tbl/vcc_types.h"), "w")
+ft = open(join(buildroot, "lib/libvcc/vcc_types.h"), "w")
 file_header(ft)
 
 lint_start(ft)
 
 for vcltype in sorted(vcltypes.keys()):
-    ft.write("VCC_TYPE(" + vcltype + ")\n")
+    ft.write("VCC_TYPE(" + vcltype + ", " + vcltype.lower() +")\n")
 ft.write("#undef VCC_TYPE\n")
 lint_end(ft)
 ft.close()
